/*----------------------------------------------------------------------
  File    : ccnpsp.c
  Contents: generate or estimate a pattern spectrum (CoCoNAD)
  Author  : Christian Borgelt
  History : 2015.08.13 file created
            2016.03.24 combined integer and graded support added
            2016.11.16 coconad miner object and interface introduced
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>
#include <float.h>
#include <time.h>
#include <assert.h>
#ifdef _WIN32
#include <windows.h>
#else
#include <unistd.h>
#include <pthread.h>
#endif
#ifdef CCN_ABORT
#include "sigint.h"
#endif
#include "random.h"
#ifndef TRN_SURR
#define TRN_SURR
#endif
#ifdef CCNPSP_MAIN
#ifndef TRN_READ
#define TRN_READ
#endif
#ifndef TA_READ
#define TA_READ
#endif
#endif
#include "train.h"
#ifndef PSP_TRAIN
#define PSP_TRAIN
#endif
#ifndef PSP_ESTIM
#define PSP_ESTIM
#endif
#ifdef CCNPSP_MAIN
#ifndef PSP_REPORT
#define PSP_REPORT
#endif
#endif
#ifndef ISR_PATSPEC
#define ISR_PATSPEC
#endif
#include "patspec.h"
#include "ccnpsp.h"
#ifdef CCNPSP_MAIN
#include "error.h"
#endif
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "ccnpsp"
#ifdef CCN_OVERLAP              /* if influence map overlap support */
#define DESCRIPTION "generate or estimate a pattern spectrum " \
                    "(CoCoNAD, overlap support)"
#else                           /* if standard (MIS) support type */
#define DESCRIPTION "generate or estimate a pattern spectrum " \
                    "(CoCoNAD)"
#endif
#define VERSION     "version 1.8 (2016.11.21)         " \
                    "(c) 2015-2016   Christian Borgelt"

/* --- thread definitions --- */
#ifdef _WIN32                   /* if Microsoft Windows system */
#define THREAD       HANDLE     /* threads identified by handles */
#define THREAD_OK    0          /* return value is DWORD */
#define WORKERDEF(n,p)  DWORD WINAPI n (LPVOID p)
#else                           /* if Linux/Unix system */
#define THREAD       pthread_t  /* use the POSIX thread type */
#define THREAD_OK    NULL       /* return value is void* */
#define WORKERDEF(n,p)  void*        n (void* p)
#endif                          /* definition of a worker function */

/* --- error codes --- */
/* error codes   0 to  -4 defined in tract.h */
#define E_STDIN      (-5)       /* double assignment of stdin */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* too few/many arguments */
#define E_TARGET     (-9)       /* invalid target type */
#define E_SIZE      (-10)       /* invalid item set size */
#define E_SUPPORT   (-11)       /* invalid item set support */
#define E_WIDTH     (-12)       /* invalid window width/max. distance */
#define E_VARIANT   (-13)       /* invalid algorithm variant */
#define E_SURRCNT   (-14)       /* invalid number of surrogates */
/* error codes -15 to -26 defined in tract.h and train.h */
#define E_SURR      (-27)       /* invalid surrogate data gen. method */
#define E_RANDFN    (-28)       /* invalid prob. density function */
#define E_ABORTED   (-29)       /* processing aborted by user */

#ifndef QUIET                   /* if not quiet version, */
#define MSG         fprintf     /* print messages */
#define XMSG        if (mode & CCN_VERBOSE) fprintf
#define CLOCK(t)    ((t) = clock())
#else                           /* if quiet version, */
#define MSG(...)    ((void)0)   /* suppress messages */
#define XMSG(...)   ((void)0)
#define CLOCK(t)    ((void)0)
#endif

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#if !defined QUIET && defined CCNPSP_MAIN
/* --- error messages --- */
static const char *errmsgs[] = {
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_TARGET   -9 */  "invalid target type '%c'",
  /* E_SIZE    -10 */  "invalid item set size %"ITEM_FMT,
  /* E_SUPPORT -11 */  "invalid minimum support %"SUPP_FMT,
  /* E_WIDTH   -12 */  "invalid window width/maximum distance %g",
  /* E_VARIANT -13 */  "invalid coconad variant '%c'",
  /* E_SURRCNT -14 */  "invalid number of surrogates %ld",
  /* E_NOITEMS -15 */  "no (frequent) items found",
  /*    -16 to -21 */  NULL, NULL, NULL, NULL, NULL, NULL,
  /*    -22 to -26 */  NULL, NULL, NULL, NULL, NULL,
  /* E_SURR    -27 */  "invalid surrogate data generation method '%c'",
  /* E_RANDFN  -28 */  "invalid random function type '%c'",
  /* E_ABORTED -29 */  "processing aborted by user",
  /*           -30 */  "unknown error"
};
#endif

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- thread worker data --- */
  int       id;                 /* id of corresponding thread */
  COCONAD   *coconad;           /* coconad miner */
  TRAINSET  *trnset;            /* trains to analyze */
  TRAINSET  *trnsur;            /* buffer for surrogate data set */
  TNSSURRFN *surrfn;            /* surrogate data generator function */
  long      cnt;                /* number of surrogate data sets */
  double    beg;                /* beginning of allowed spike range */
  double    end;                /* end       of allowed spike range */
  double    delta;              /* block size for blocked perms. */
  RNG       *rng;               /* random number generator */
  ISREPORT  *report;            /* item set reporter (one per thread) */
  int       err;                /* error indicator */
  volatile long *done;          /* number of completed data sets */
  PRGREPFN  *repfn;             /* progress reporting function */
  void      *data;              /* progress reporting function data */
} WORKDATA;                     /* (thread worker data) */

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static TNSSURRFN *sur_tab[] = { /* surrogate generation functions */
  /* IDENTITY 0 */ tns_ident,   /* identity (keep original data) */
  /* RANDOM   1 */ tns_random,  /* event/spike time randomization */
  /* DITHER   2 */ tns_dither,  /* event/spike time dithering */
  /* SHIFT    3 */ tns_shift,   /* train shifting/dithering */
  /* KERNEST  4 */ tns_kernest, /* sampling from kernel estimate */
  /* PERMUTE  5 */ tns_permute, /* dithered spike permutation */
  /* BLOCK    6 */ tns_permblk, /* dithered blocked permutations */
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
#ifdef CCNPSP_MAIN
#ifndef QUIET
static CCHAR    *prgname;       /* program name for error messages */
#endif
static TABREAD  *tread  = NULL; /* table/train reader */
static ITEMBASE *ibase  = NULL; /* item base */
static TRAINSET *trnset = NULL; /* train set */
static PATSPEC  *psp    = NULL; /* pattern spectrum */
static TABWRITE *twrite = NULL; /* table writer */
static RSUPP    *border = NULL; /* decision border */
#endif

/*----------------------------------------------------------------------
  CPUinfo Function
----------------------------------------------------------------------*/

static int cpucnt (void)
{                               /* --- get the number of processors */
  #ifdef _WIN32                 /* if Microsoft Windows system */
  SYSTEM_INFO sysinfo;          /* system information structure */
  GetSystemInfo(&sysinfo);      /* get system information */
  return (int)sysinfo.dwNumberOfProcessors;
  #elif defined _SC_NPROCESSORS_ONLN
  return (int)sysconf(_SC_NPROCESSORS_ONLN);
  #else                         /* if no direct function available */
  static int cpuinfo[5] = { 0, 0, 0, 0, 0 };
  if (!cpuinfo[4]) {            /* if cpu info. not yet retrieved */
    __asm__ __volatile__ (      /* execute cpuid instruction */
      "cpuid": "=a" (cpuinfo[0]), "=b" (cpuinfo[1]),
               "=c" (cpuinfo[2]), "=d" (cpuinfo[3]) : "a" (1) );
    cpuinfo[4] = -1;            /* execute cpuid instruction */
  }                             /* and set initialization flag */
  return (cpuinfo[1] >> 16) & 0xff;  /* get the number of */
  #endif                        /* addressable processors */
}  /* cpucnt() */

/*----------------------------------------------------------------------
  Pattern Spectrum Generation
----------------------------------------------------------------------*/

static WORKERDEF(worker, p)
{                               /* --- worker function for a thread */
  WORKDATA *w = p;              /* type the argument pointer */
  long     i;                   /* loop variable for data sets */

  assert(p);                    /* check the function argument */
  for (i = 1; i <= w->cnt; i++){/* generate cnt surrogate data sets */
    w->trnsur = w->surrfn(w->trnset, w->beg, w->end,
                          w->delta,  w->rng, w->trnsur);
    #ifdef CCN_ABORT            /* if a signal handler is present */
    if (sig_aborted()) break;   /* check for an abort interrupt */
    #endif
    w->err |= ccn_data(w->coconad, w->trnsur, CCN_SURR);
    if (w->err < 0)    break;   /* (re-)prepare the surrogate data */
    #ifdef CCN_ABORT            /* if a signal handler is present */
    if (sig_aborted()) break;   /* check for an abort interrupt */
    #endif
    w->err |= ccn_mine(w->coconad);
    if (w->err < 0)    break;   /* execute the coconad algorithm */
    #ifdef CCN_ABORT            /* if a signal handler is present */
    if (sig_aborted()) break;   /* check for an abort interrupt */
    #endif
    w->done[0]++;               /* count the surrogate data set */
    if (w->repfn) w->repfn(w->done[0], w->data);
  }                             /* report the progress */
  return THREAD_OK;             /* return a dummy result */
}  /* worker() */

/*--------------------------------------------------------------------*/

PATSPEC* ccn_genpsp (TRAINSET *trnset, int target, double width,
                     double supp,
#ifdef CCN_OVERLAP              /* if influence map overlap support */
                     double mix,
#endif                          /* additional mixing parameter */
                     ITEM zmin, ITEM zmax,
                     int algo, int mode, size_t cnt,
                     double beg, double end, int surr, int rand,
                     double sigma, double delta, long seed,
                     int cpus, PRGREPFN *rep, void* data)
{                               /* --- generate a pattern spectrum */
  PATSPEC   *psp = NULL;        /* created pattern spectrum */
  int       r;                  /* result of function call */
  ITEMBASE  *base;              /* underlying item base */
  TRAINSET  *trnsur, *s;        /* surrogate data set */
  TNSSURRFN *surrfn;            /* surrogate data generation function */
  RNG       *rng;               /* random number generator */
  ISREPORT  *report;            /* item set reporter */
  COCONAD   *coconad;           /* coconad miner */
  THREAD    *threads;           /* thread handles */
  WORKDATA  *w;                 /* data for worker thread */
  long      c, x;               /* number of data sets per thread */
  long      i;                  /* loop variable for data sets */
  int       k, n;               /* loop variables for threads */
  #ifdef _WIN32                 /* if Microsoft Windows system */
  DWORD     thid;               /* dummy for storing the thread id */
  #endif                        /* (not really needed here) */
  volatile long done = 0;       /* number of completed surrogates */

  assert(trnset                 /* check the function arguments */
  &&    (algo >= CCN_BASIC) && (algo <= CCN_AUTO));
  if (seed == 0) seed = (long)time(NULL);

  /* --- prepare data --- */
  supp = ceilsupp(supp);        /* get absolute minimum support */
  #ifndef CCN_OVERLAP           /* if standard (MIS) support */
  coconad = ccn_create(target, width, (SUPP)supp,
                       zmin, zmax, algo, mode);
  #else                         /* if influence map overlap support */
  coconad = ccn_create(target, width, (SUPP)supp, mix,
                       zmin, zmax, algo, mode);
  #endif                        /* add. support mixing parameter */
  if (!coconad) return NULL;    /* create a coconad miner */
  if ((beg <= -INFINITY) || (beg >= INFINITY))
    beg = floor(tns_min(trnset, 0)); /* get start of event span */
  if ((end <= -INFINITY) || (end >= INFINITY))
    end = ceil (tns_max(trnset, 0)); /* get end   of event span */
  tns_filter(trnset, beg, end);      /* filter the train set */
  r = ccn_data(coconad, trnset, 0);  /* prepare the data for coconad */
  if (r) { ccn_delete(coconad, 0); return NULL; }

  /* --- generate pattern spectrum --- */
  base = tns_base(trnset);      /* get underlying item base */
  if (cpus <= 0) cpus = cpucnt();
  if ((cpus > 1) && (cnt > 1)){ /* if to use multi-threading */
    threads = calloc((size_t)cpus, sizeof(THREAD));
    if (!threads) return NULL;  /* create array of thread handles */
    w = calloc((size_t)cpus, sizeof(WORKDATA));
    if (!w) { free(threads); return NULL; }
    c = ((long)cnt+cpus-1)/cpus;/* number of data sets per thread */
    for (n = 0; n < cpus; n++){ /* traverse the cpus/threads */
      x = (long)cnt -n*c;       /* get the number of data sets and */
      if (x <= 0) break;        /* check whether thread is needed */
      if (!coconad) {           /* if no miner for this thread */
        #ifndef CCN_OVERLAP     /* if standard (MIS) support */
        coconad = ccn_create(target, width, (SUPP)supp,
                             zmin, zmax, algo, mode);
        #else                   /* if influence map overlap support */
        coconad = ccn_create(target, width, (SUPP)supp, mix,
                             zmin, zmax, algo, mode);
        #endif                  /* add. support mixing parameter */
      }                         /* create a coconad miner */
      w[n].id      = n;         /* note the thread identifier */
      w[n].coconad = coconad;   /* note the coconad miner and */
      w[n].trnset  = trnset;    /* the train set and its clone */
      w[n].trnsur  = tns_clone(trnset);
      w[n].surrfn  = sur_tab[surr];
      w[n].cnt     = (x < c) ? x : c;
      w[n].beg     = beg;       /* note beginning and end */
      w[n].end     = end;       /* of allowed event region */
      w[n].delta   = delta;     /* note the block size */
      w[n].rng     = rng_createx((unsigned int)(seed+n), rand, sigma);
      w[n].report  = isr_create(base);
      w[n].err     = 0;         /* create random number generator */
      w[n].done    = &done;     /* and an item set reporter */
      w[n].repfn   = rep;
      w[n].data    = data;
      if (!w[n].coconad || !w[n].trnsur || !w[n].rng || !w[n].report) {
        w[n].err = -1; break; } /* check for successful creation */
      if ((ccn_report(w[n].coconad, w[n].report) != 0)
      ||  (isr_addpsp(w[n].report, NULL) < 0)
      ||  (isr_setup (w[n].report) != 0)) {
        w[n].err = -1; break; } /* set up miner and item set reporter */
      #ifdef _WIN32             /* if Microsoft Windows system */
      threads[n] = CreateThread(NULL, 0, worker, w+n, 0, &thid);
      if (!threads[n]) { w[n].err = -1; break; }
      #else                     /* if Linux/Unix system */
      if (pthread_create(threads+n, NULL, worker, w+n) != 0) {
        w[n].err = -1; break; } /* create a thread for each data set */
      #endif                    /* to compute surrogates in parallel */
      coconad = NULL;           /* current miner has been used */
    }
    #ifdef _WIN32               /* if Microsoft Windows system */
    WaitForMultipleObjects((DWORD)n, threads, TRUE, INFINITE);
    for (k = n; --k >= 0; )     /* wait for threads to finish, */
      CloseHandle(threads[k]);  /* then close all thread handles */
    #else                       /* if Linux/Unix system */
    for (k = n; --k >= 0; )     /* wait for threads to finish */
      pthread_join(threads[k], NULL);
    #endif                      /* (join threads with this one) */
    for (k = r = 0; k < n; k++)
      r |= w[k].err;            /* join the error indicators */
    if (r >= 0) {               /* if processing was successful */
      psp = isr_rempsp(w[0].report, 0);
      for (k = 1; k < n; k++) { /* traverse the threads */
        r = psp_addpsp(psp, isr_getpsp(w[k].report));
        if (r < 0) break;       /* traverse and sum pattern spectrums */
      }                         /* in the one of the first thread */
    }
    for (k = n; --k >= 0;) {    /* traverse the worker data */
      if (w[k].coconad) ccn_delete(w[k].coconad, 0);
      if (w[k].trnsur)  tns_delete(w[k].trnsur, 0);
      if (w[k].rng)     rng_delete(w[k].rng);
      if (w[k].report)  isr_delete(w[k].report,  0);
    }                           /* delete the data structures */
    free(w);                    /* delete array of worker data */
    free(threads); }            /* and array of thread handles */
  else {                        /* if to use only one thread */
    report = isr_create(base);  /* create an item set reporter */
    if (!report) { ccn_delete(coconad, 0); return NULL; }
    if ((ccn_report(coconad, report) < 0)
    ||  (isr_addpsp(report, NULL) < 0)
    ||  (isr_setup (report) != 0)) {
      isr_delete(report, 0); ccn_delete(coconad, 0); return NULL; }
    rng = rng_createx((unsigned)seed, rand, sigma);
    if (!rng) {                 /* create a random number generator */
      isr_delete(report, 0); ccn_delete(coconad, 0); return NULL; }
    surrfn = sur_tab[surr];     /* get the surrogate data function */
    trnsur = NULL; r = 0;       /* init. surrogate and return code */
    for (i = 1; i <= (long)cnt; i++) {
      s = surrfn(trnset, beg, end, delta, rng, trnsur);
      if (!s) { r = -1; break;} /* generate a surrogate data set */
      r = ccn_data(coconad, trnsur = s, CCN_SURR);
      if (r < 0) break;         /* prepare the data set and */
      r = ccn_mine(coconad);    /* execute the coconad algorithm */
      if (r < 0) break;
      if (rep) rep(i, data);    /* report the progress and */
      #ifdef CCN_ABORT          /* if a signal handler is present */
      if (sig_aborted()) break; /* check for an interrupt */
      #endif
    }
    psp = isr_rempsp(report,0); /* get the created pattern spectrum */
    if (trnsur) tns_delete(trnsur, 0);
    ccn_delete(coconad, 0);     /* delete the data objects */
    rng_delete(rng);            /* (coconad miner, */
    isr_delete(report, 0);      /* random number generator, */
  }                             /* and item set reporter) */

  /* --- clean up --- */
  #ifdef CCN_ABORT              /* if a signal handler is present */
  if (sig_aborted()) { sig_remove(); return NULL; }
  #endif
  if (r < 0) { if (psp) psp_delete(psp); return NULL; }
  #ifdef CCN_ABORT              /* if a signal handler is present */
  sig_remove();                 /* remove the signal handler */
  #endif
  return psp;                   /* return the created Java object */
}  /* ccn_genpsp() */

/*----------------------------------------------------------------------
  Pattern Spectrum Estimation
----------------------------------------------------------------------*/

PATSPEC* ccn_estpsp (TRAINSET *trnset, int target, double width,
                     double supp, ITEM zmin, ITEM zmax, size_t equiv,
                     double alpha, size_t smpls, long seed)
{                               /* --- estimate a pattern spectrum */
  PATSPEC *psp;                 /* created pattern spectrum */

  assert(trnset);               /* check the function arguments */
  if (seed <= 0) seed = (long)time(NULL);
  rseed((unsigned)seed);        /* seed random number generator */
  supp = ceilsupp(supp);        /* get absolute minimum support */
  psp  = psp_create(zmin, zmax, (SUPP)supp, tns_maxlen(trnset));
  if (!psp) return NULL;        /* create a pattern spectrum */
  if      (target & CCN_SEQUENCE) target = PSP_SEQUENCE;
  else if (target & CCN_PERMUT)   target = PSP_PERMUT;
  else                            target = PSP_ITEMSET;
  if (psp_tnsest(trnset, psp, (size_t)equiv, width,
                 alpha, (size_t)smpls, target) == 0)
    return psp;                 /* estimate pattern spectrum */
  psp_delete(psp); return NULL; /* on failure delete pattern spectrum */
}  /* ccn_estpsp() */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/
#ifdef CCNPSP_MAIN

static void help (void)
{                               /* --- print add. option information */
  #ifndef QUIET
  fprintf(stderr, "\n");        /* terminate startup message */
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  printf("coconad algorithm variants (option -A#, default: -Ai)\n");
  #else                         /* if standard (MIS) support */
  printf("coconad algorithm variants (option -A#, default: -Ar/f)\n");
  #endif                        /* (different default algorithms) */
  printf("  b   basic version with linear memory consumption\n");
  printf("  f   refined version with event and train filtering\n");
  printf("  r   closed/maximal patterns filtered with a repository\n");
  printf("      (equivalent to variant 'f' for frequent patterns)\n");
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  printf("  i   interval list intersection (c/m with repository)\n");
  #else                         /* special algorithm variant */
  printf("(default is 'f' with options -Tp/s, otherwise it is 'r')\n");
  #endif                        /* special algorithm variant */
  printf("surrogate data generation methods "
         "(option -g#, default: -ge)\n");
  printf("  e    estimate a pattern spectrum (no surrogates)\n");
  printf("  i    identity (keep original data)\n");
  printf("  r    event/spike time randomization\n");
  printf("  d    event/spike dithering/displacement\n");
  printf("  s    train shifting/dithering\n");
  printf("  k    sampling from a kernel estimate\n");
  printf("  p    dithered item permutation\n");
  printf("  b    dithered blocked item permutations\n");
  printf("random function type for displacements "
         "(option -d#, default: -gr)\n");
  printf("  u/r  uniform/rectangular density\n");
  printf("  t    symmetric triangular density\n");
  printf("  g/n  Gaussian/normal density\n");
  #endif                        /* print help information */
  exit(0);                      /* abort the program */
}  /* help() */

/*--------------------------------------------------------------------*/

static void repfn (long cnt, void *data)
{                               /* --- progress reporting function */
  if ((cnt > *(long*)data) && ((cnt % 20) == 0))
    fprintf(stderr, "%10ld\b\b\b\b\b\b\b\b\b\b", *(long*)data = cnt);
}  /* repfn() */

/*--------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (twrite) twr_delete(twrite, 1); \
  if (psp)    psp_delete(psp);       \
  if (trnset) tns_delete(trnset, 0); \
  if (tread)  trd_delete(tread,  1); \
  if (ibase)  ib_delete (ibase);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counters */
  char    *s;                   /* to traverse the options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_inp  = NULL;      /* name of input  file */
  CCHAR   *fn_sel  = NULL;      /* name of item selection file */
  CCHAR   *fn_psp  = NULL;      /* name of pattern spectrum file */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *comment = NULL;      /* comment characters */
  int     target   = 'i';       /* target type (sets/permutations) */
  int     trgsub   = 's';       /* target subtype (closed/maximal) */
  double  width    = 0.003;     /* window width/maximum distance */
  SUPP    smin     = 2;         /* minimum support of an item set */
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  double  mix      = 0.0;       /* mixing coefficient for standard */
  #endif                        /* and influence map overlap support */
  ITEM    zmin     = 2;         /* minimum size of an item set */
  ITEM    zmax     = ITEM_MAX;  /* maximum size of an item set */
  int     algo     = 0;         /* variant of coconad algorithm */
  int     mode     = CCN_DEFAULT;  /* search mode (e.g. pruning) */
  long    cnt      = 1000;      /* number of surrogate data sets */
  double  beg      = -INFINITY; /* beginning of event range */
  double  end      =  INFINITY; /* end       of event range */
  int     surr     = 'e';       /* surrogate data generation method */
  int     rand     = 'r';       /* probability density function */
  double  sigma    = 0.005;     /* parameter of random function */
  double  delta    = 0.03;      /* limit for localized swaps */
  double  alpha    = 0.5;       /* probability dispersion factor */
  long    smpls    = 1000;      /* number of item set samples */
  long    seed     = 0;         /* seed for random numbers */
  int     cpus     = 0;         /* number of cpus to use */
  long    done     = 0;         /* number of completed data sets */
  int     minp     = TRN_ITEMPT;/* mode for train set reading */
  int     decbdr   = 0;         /* flag for printing the border */
  ITEM    m;                    /* number of items/trains */
  size_t  n;                    /* number of events/spikes */
  RSUPP   bmin;                 /* for printing the decision border */
  #ifndef QUIET                 /* if not quiet version */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no arguments given */
    printf("usage: %s [options] infile [outfile]\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    #ifndef CCN_OVERLAP         /* if standard (MIS) support type */
    printf("-T#      target pattern type                      "
                    "(default: %c)\n", target);
    printf("         (i: item sets, p: partial permutations "
                    "(without repetition),\n");
    printf("          s: general sequences "
                    "(with & without repetition))\n");
    printf("-t#      target pattern subtype                   "
    #else                       /* if influence map overlap support */
    printf("-t#      target pattern type                      "
    #endif
                    "(default: %c)\n", trgsub);
    printf("         (s/f: frequent, c: closed, "
                     "m: maximal patterns)\n");
    printf("-m#      minimum number of items per item set     "
                    "(default: %"ITEM_FMT")\n", zmin);
    printf("-n#      maximum number of items per item set     "
                    "(default: no limit)\n");
    printf("-w#      width of time window/maximum distance    "
                    "(default: %g)\n", width);
    printf("-s#      minimum support of an item set           "
                    "(default: %"SUPP_FMT")\n", smin);
    #ifdef CCN_OVERLAP          /* if influence map overlap support */
    printf("-M#      mixing coefficient for MIS support       "
                    "(default: %g)\n", mix);
    #endif                      /* sequence mining is possible */
    printf("-A#      variant of the algorithm to use          "
                    "(default: r/f)\n");
    printf("         (default is f with option -o and r without)\n");
    printf("-0       list/array for the repository            "
                    "(default: tree chain)\n");
    #ifndef CCN_OVERLAP         /* if standard (MIS) support type */
    printf("-1       single prefix tree for the repository    "
                    "(default: tree chain)\n");
    printf("         (only for partial permutations and sequences)\n");
    printf("-p       do not sort trains by their length       "
                    "(default: sort)\n");
    #endif                      /* sorting by train length may help */
    printf("-x       do not prune with perfect extensions     "
                    "(default: prune)\n");
    printf("-a#      beginning of event/spike range           "
                    "(default: -oo)\n");
    printf("-z#      end       of event/spike range           "
                    "(default: +oo)\n");
    printf("-g#      surrogate generation method              "
                    "(default: %c)\n", surr);
    printf("-c#      number of surrogate data sets            "
                    "(default: %ld)\n", cnt);
    printf("-d#      random function type for displacements   "
                    "(default: %c)\n", rand);
    printf("-j#      parameter for random function density    "
                    "(default: %g)\n", sigma);
    printf("         (standard deviation or half base width "
                    "of rectangle/triangle)\n");
    printf("-B#      block size for block permutations        "
                    "(default: %g)\n", delta);
    printf("         (only used with option -gb: "
                    "block permutations)\n");
    printf("-e#      probability dispersion factor            "
                    "(default: %g)\n",  alpha);
    printf("-i#      number of item set samples per size      "
                    "(default: %ld)\n", smpls);
    printf("-S#      seed for random numbers                  "
                    "(default: time)\n");
    printf("-Z#      number of cpus/processor cores to use    "
                    "(default: %d)\n", cpus);
    printf("         (a value <= 0 means all cpus "
                    "reported as available)\n");
    printf("-D       print induced decision border\n");
    printf("-l       one train per record (item + events)     "
                    "(default: id/time pairs)\n");
    printf("-y       no item (with -l) or item after time     "
                    "(default: item first)\n");
    printf("-R#      read an item selection from a file       "
                    "(default: use all items)\n");
    printf("-r#      record/transaction separators            "
                    "(default: \"\\n\")\n");
    printf("-f#      field /item        separators            "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                       "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters                       "
                    "(default: \"#\")\n");
    printf("-!       print additional option information\n");
    printf("infile   file to read trains from                 "
                    "[required]\n");
    printf("outfile  file to write pattern spectrum to        "
                    "[optional]\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */
  #endif  /* #ifndef QUIET */

  /* --- evaluate arguments --- */
  seed = (long)time(NULL);      /* and get a default seed value */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (*s) {              /* traverse options */
        switch (*s++) {         /* evaluate switches */
          case '!': help();                          break;
          #ifndef CCN_OVERLAP   /* if standard (MIS) support */
          case 'T': target = (*s) ? *s++ : 's';      break;
          #endif
          case 't': trgsub = (*s) ? *s++ : 's';      break;
          case 'm': zmin   = (ITEM)strtol(s, &s, 0); break;
          case 'n': zmax   = (ITEM)strtol(s, &s, 0); break;
          case 'w': width  =       strtod(s, &s);    break;
          case 's': smin   =    strtosupp(s, &s);    break;
          #ifdef CCN_OVERLAP    /* if influence map overlap support */
          case 'M': mix    =       strtod(s, &s);    break;
          #endif                /* (support mixing parameter) */
          case 'A': algo   = (*s) ? *s++ : 0;        break;
          #ifndef CCN_OVERLAP   /* if standard (MIS) support */
          case '1': mode  |=  CCN_REPOTREE;          break;
          #endif                /* (single prefix tree repository) */
          case '0': mode  |=  CCN_REPOLIST;          break;
          case 'x': mode  &= ~CCN_PERFECT;           break;
          case 'p': mode  &= ~CCN_REORDER;           break;
          case 'a': beg    =       strtod(s, &s);    break;
          case 'z': end    =       strtod(s, &s);    break;
          case 'g': surr   = (*s) ? *s++ : 0;        break;
          case 'c': cnt    =       strtol(s, &s, 0); break;
          case 'd': rand   = (*s) ? *s++ : 0;        break;
          case 'j': sigma  =       strtod(s, &s);    break;
          case 'B': delta  =       strtod(s, &s);    break;
          case 'e': alpha  =       strtod(s, &s);    break;
          case 'i': smpls  =       strtol(s, &s, 0); break;
          case 'S': seed   =       strtol(s, &s, 0); break;
          case 'Z': cpus   =  (int)strtol(s, &s, 0); break;
          case 'D': decbdr = -1;                     break;
          case 'l': minp  &= ~TRN_ITEMPT;            break;
          case 'y': minp  |=  TRN_PURETRN;           break;
          case 'R': optarg = &fn_sel;                break;
          case 'r': optarg = &recseps;               break;
          case 'f': optarg = &fldseps;               break;
          case 'b': optarg = &blanks;                break;
          case 'C': optarg = &comment;               break;
          default : error(E_OPTION, *--s);           break;
        }                       /* set option variables */
        if (optarg && *s) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-options */
        case  0: fn_inp = s;      break;
        case  1: fn_psp = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check (option) arguments */
  if (k < 2)  error(E_ARGCNT);  /* and number of arguments */
  if (width <= 0) error(E_WIDTH,   width);
  if (smin  == 0) error(E_SUPPORT, smin);
  if (smin  <  0) smin = -smin; /* ensure positive minimum support */
  if (zmin  <  1) error(E_SIZE, zmin); /* check the size limits */
  if (zmax  <  1) error(E_SIZE, zmax);
  if (cnt   <= 0) error(E_SURRCNT, cnt);
  switch (target) {             /* check and translate target type */
    case 'i': target = CCN_ITEMSET;          break;
    case 'p': target = CCN_PERMUT;           break;
    case 's': target = CCN_SEQUENCE;         break;
    default : error(E_TARGET, (char)target); break;
  }                             /* (get target type code) */
  switch (trgsub) {             /* check and translate target subtype */
    case 's': target |= CCN_FREQUENT;        break;
    case 'f': target |= CCN_FREQUENT;        break;
    case 'c': target |= CCN_CLOSED;          break;
    case 'm': target |= CCN_MAXIMAL;         break;
    default : error(E_TARGET, (char)trgsub); break;
  }                             /* (get target type code) */
  if (algo == 0)                /* choose default algorithm */
    algo = (target & (CCN_PERMUT|CCN_SEQUENCE)) ? 'f' : 'r';
  switch (algo) {               /* check and translate alg. variant */
    case 'b': algo = CCN_BASIC;              break;
    case 'f': algo = CCN_FILTER;             break;
    case 'r': algo = CCN_REPO;               break;
    #ifdef CCN_OVERLAP          /* if influence map overlap support */
    case 'i': algo = CCN_ISECT;              break;
    #endif                      /* additional coconad variant(s) */
    default : error(E_VARIANT, (char)algo);  break;
  }                             /* (get coconad algorithm code) */
  switch (surr) {               /* check and translate surrogate */
    case 'e': surr = -1;                     break;
    case 'i': surr = CCN_IDENTITY;           break;
    case 'r': surr = CCN_RANDOM;             break;
    case 'd': surr = CCN_DITHER;             break;
    case 's': surr = CCN_SHIFT;              break;
    case 'k': surr = CCN_KERNEST;            break;
    case 'p': surr = CCN_PERMUTE;            break;
    case 'b': surr = CCN_BLOCK;              break;
    default : error(E_SURR, (char)surr);     break;
  }                             /* (get surrogate method code) */
  if (surr == CCN_IDENTITY) cnt = 1;
  switch (rand) {               /* check and translate random fn. */
    case 'u': rand = RNG_UNIFORM;            break;
    case 'r': rand = RNG_RECT;               break;
    case 't': rand = RNG_TRIANG;             break;
    case 'g': rand = RNG_GAUSS;              break;
    case 'n': rand = RNG_NORMAL;             break;
    default : error(E_RANDFN, (char)rand);   break;
  }                             /* (get random function code) */
  MSG(stderr, "\n");            /* terminate the startup message */

  /* --- read item selection --- */
  ibase = ib_create(0, 0);      /* create an item base */
  if (!ibase) error(E_NOMEM);   /* to manage the items */
  tread = trd_create();         /* create a transaction reader */
  if (!tread) error(E_NOMEM);   /* and configure the characters */
  trd_allchs(tread, recseps, fldseps, blanks, "", comment);
  if (fn_sel) {                 /* if item appearances are given */
    CLOCK(t);                   /* start timer, open input file */
    if (trd_open(tread, NULL, fn_sel) != 0)
      error(E_FOPEN, trd_name(tread));
    MSG(stderr, "reading %s ... ", trd_name(tread));
    m = ib_readsel(ibase,tread);/* read the given item selection */
    if (m < 0) error((int)-m, ib_errmsg(ibase, NULL, 0));
    trd_close(tread);           /* close the input file */
    MSG(stderr, "[%"ITEM_FMT" item(s)]", m);
    MSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }                             /* print a log message */

  /* --- read train set --- */
  trnset = tns_create(ibase);   /* create a train set */
  if (!trnset) error(E_NOMEM);  /* to store the trains */
  CLOCK(t);                     /* start timer, open input file */
  if (trd_open(tread, NULL, fn_inp) != 0)
    error(E_FOPEN, trd_name(tread));
  MSG(stderr, "reading %s ... ", trd_name(tread));
  k = tns_read(trnset, tread, minp);
  if (k < 0)                    /* read the transaction database */
    error(-k, tns_errmsg(trnset, NULL, 0));
  trd_delete(tread, 1);         /* close the input file and */
  tread = NULL;                 /* delete the table reader */
  m = tns_cnt(trnset);          /* get the number of trains */
  n = tns_extent(trnset);       /* and the number of events */
  MSG(stderr, "[%"ITEM_FMT" train(s), %"SIZE_FMT" event(s)]", m, n);
  MSG(stderr, " done [%.2fs].", SEC_SINCE(t));
  if ((m <= 0) || (n <= 0))     /* check for at least one item */
    error(E_NOITEMS);           /* and at least one event/spike */
  MSG(stderr, "\n");            /* terminate the log message */

  /* --- sort train set --- */
  CLOCK(t);                     /* start timer, print log message */
  MSG(stderr, "sorting events in trains ... ");
  tns_sort(trnset, +1);         /* sort the trains by events */
  MSG(stderr, "[%"ITEM_FMT" train(s), %"SIZE_FMT" event(s)]", m, n);
  MSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- estimate/generate pattern spectrum --- */
  CLOCK(t);                     /* start timer, print log message */
  #ifndef CCN_OVERLAP           /* if standard (MIS) support */
  if (surr < 0) {               /* if to estimate a pattern spectrum */
    MSG(stderr, "estimating pattern spectrum ... ");
    psp = ccn_estpsp(trnset, target, width, smin, zmin, zmax,
                     (size_t)cnt, alpha, (size_t)smpls, seed); }
  else {                        /* if to generate a pattern spectrum */
    MSG(stderr, "generating pattern spectrum ... ");
    psp = ccn_genpsp(trnset, target, width, smin, zmin, zmax,
                     algo, mode, (size_t)cnt, beg, end, surr, rand,
                     sigma, delta, seed, cpus, repfn, &done);
  }                             /* (create a pattern spectrum) */
  #else                         /* if influence map overlap support */
  MSG(stderr, "generating pattern spectrum ... ");
  psp = ccn_genpsp(trnset, target, width, smin, mix, zmin, zmax,
                   algo, mode, (size_t)cnt, beg, end, surr, rand,
                   sigma, delta, seed, cpus, repfn, &done);
  #endif
  #ifdef CCN_ABORT              /* if a signal handler is present */
  if (!psp) error((sig_aborted()) ? E_ABORTED : E_NOMEM);
  #else                         /* check for abort or error */
  if (!psp) error(E_NOMEM);     /* if no signal handler is present */
  #endif                        /* it must be out of memory */
  n = psp_sigcnt(psp);          /* get the number of signatures */
  MSG(stderr, "[%"SIZE_FMT" signature(s)]", n);
  MSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- write pattern spectrum --- */
  CLOCK(t);                     /* start timer, print log message */
  twrite = twr_create();        /* create a table writer and */
  if (!twrite) error(E_NOMEM);  /* open the output file */
  if (twr_open(twrite, NULL, fn_psp) != 0)
    error(E_FOPEN,  twr_name(twrite));
  MSG(stderr, "writing %s ... ", twr_name(twrite));
  if (psp_report(psp, twrite, 1.0/(double)cnt) != 0)
    error(E_FWRITE, twr_name(twrite));
  twr_delete(twrite, 1);        /* write the pattern spectrum */
  twrite = NULL;                /* and delete the table writer */
  MSG(stderr, "[%"SIZE_FMT" signature(s)]", n);
  MSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- print decision border --- */
  if (decbdr) {                 /* if to print the decision border */
    for (zmax = psp_max(psp); zmax >= zmin; zmax--)
      if (psp_max4sz(psp, zmax) >= psp_min4sz(psp, zmin))
        break;                  /* find the largest pattern size */
    bmin   = (RSUPP)smin;       /* get the minimum support */
    border = (RSUPP*)malloc(((size_t)zmax+1) *sizeof(RSUPP));
    if (!border) error(E_NOMEM);/* create a decision border */
    for (m = zmax+1; --m >= zmin; ) { /* traverse the pattern sizes */
      border[m] = psp_max4sz(psp, m) +SUPP_EPS;
      if (bmin > border[m]) border[m] = bmin;
      bmin = border[m];         /* get min. support for significance */
    }                           /* and ensure a monotone border */
    for (m = zmin; m < zmax; m++) printf("%"RSUPP_FMT":", border[m]);
    if (zmax > zmin) { printf("%"RSUPP_FMT"\n", border[zmax]); }
  }                             /* print the decision border */

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */

#endif  /* #ifdef CCNPSP_MAIN */
